/*
 * Copyright 2018-2024 KMath contributors.
 * Use of this source code is governed by the Apache 2.0 license that can be found in the license/LICENSE.txt file.
 */


@file:OptIn(PerformancePitfall::class)

package space.kscience.kmath.tensors.core

import space.kscience.kmath.PerformancePitfall
import space.kscience.kmath.nd.*
import space.kscience.kmath.operations.Float64BufferOps
import space.kscience.kmath.operations.Float64Field
import space.kscience.kmath.structures.*
import space.kscience.kmath.tensors.api.AnalyticTensorAlgebra
import space.kscience.kmath.tensors.api.LinearOpsTensorAlgebra
import space.kscience.kmath.tensors.api.Tensor
import space.kscience.kmath.tensors.core.internal.*
import kotlin.math.abs
import kotlin.math.ceil
import kotlin.math.floor
import kotlin.math.sqrt

/**
 * Implementation of basic operations over double tensors and basic algebra operations on them.
 */
@OptIn(PerformancePitfall::class)
public open class DoubleTensorAlgebra :
    AnalyticTensorAlgebra<Double, Float64Field>,
    LinearOpsTensorAlgebra<Double, Float64Field> {

    public companion object : DoubleTensorAlgebra()

    override val elementAlgebra: Float64Field get() = Float64Field

    public val bufferAlgebra: Float64BufferOps get() = Float64BufferOps


    /**
     * Applies the [transform] function to each element of the tensor and returns the resulting modified tensor.
     *
     * @param transform the function to be applied to each element of the tensor.
     * @return the resulting tensor after applying the function.
     */
    @Suppress("OVERRIDE_BY_INLINE")
    final override inline fun StructureND<Double>.map(transform: Float64Field.(Double) -> Double): DoubleTensor {
        val tensor = asDoubleTensor()
        //TODO remove additional copy
        val array = Float64Buffer(tensor.source.size) { Float64Field.transform(tensor.source[it]) }
        return DoubleTensor(
            tensor.shape,
            array,
        )
    }

    public inline fun Tensor<Double>.mapInPlace(operation: (Double) -> Double) {
        if (this is DoubleTensor) {
            source.mapInPlace(operation)
        } else {
            indices.forEach { set(it, operation(get(it))) }
        }
    }

    public inline fun Tensor<Double>.mapIndexedInPlace(operation: Float64Field.(IntArray, Double) -> Double) {
        indices.forEach { set(it, Float64Field.operation(it, get(it))) }
    }

    @Suppress("OVERRIDE_BY_INLINE")
    final override inline fun StructureND<Double>.mapIndexed(transform: Float64Field.(index: IntArray, Double) -> Double): DoubleTensor {
        return copyToTensor().apply { mapIndexedInPlace(transform) }
    }


    @Suppress("OVERRIDE_BY_INLINE")
    final override inline fun zip(
        left: StructureND<Double>,
        right: StructureND<Double>,
        transform: Float64Field.(Double, Double) -> Double,
    ): DoubleTensor {
        checkShapesCompatible(left, right)

        val leftTensor = left.asDoubleTensor()
        val rightTensor = right.asDoubleTensor()
        val buffer = Float64Buffer(leftTensor.source.size) {
            Float64Field.transform(leftTensor.source[it], rightTensor.source[it])
        }
        return DoubleTensor(leftTensor.shape, buffer)
    }


    public inline fun StructureND<Double>.reduceElements(transform: (Float64Buffer) -> Double): Double =
        transform(asDoubleTensor().source.copy())
    //TODO Add read-only DoubleBuffer wrapper. To avoid protective copy

    override fun StructureND<Double>.valueOrNull(): Double? {
        val dt = asDoubleTensor()
        return if (dt.shape contentEquals ShapeND(1)) dt.source[0] else null
    }

    override fun StructureND<Double>.value(): Double = valueOrNull()
        ?: throw IllegalArgumentException("The tensor shape is $shape, but value method is allowed only for shape [1]")

    public fun fromBuffer(shape: ShapeND, buffer: Buffer<Double>): DoubleTensor {
        checkNotEmptyShape(shape)
        check(buffer.size > 0) { "Illegal empty buffer provided" }
        check(buffer.size == shape.linearSize) {
            "Inconsistent shape $shape for buffer of size ${buffer.size} provided"
        }
        return DoubleTensor(shape, buffer.toFloat64Buffer())
    }


    /**
     * Constructs a tensor with the specified shape and data.
     *
     * @param shape the desired shape for the tensor.
     * @param array one-dimensional data array.
     * @return tensor with the [shape] shape and [array] data.
     */
    public fun fromArray(shape: ShapeND, array: DoubleArray): DoubleTensor = fromBuffer(shape, array.asBuffer())

    /**
     * Constructs a tensor with the specified shape and initializer.
     *
     * @param shape the desired shape for the tensor.
     * @param initializer mapping tensor indices to values.
     * @return tensor with the [shape] shape and data generated by the [initializer].
     */
    override fun mutableStructureND(shape: ShapeND, initializer: Float64Field.(IntArray) -> Double): DoubleTensor =
        fromArray(
            shape,
            RowStrides(shape).asSequence().map { Float64Field.initializer(it) }.toMutableList().toDoubleArray()
        )

    override fun Tensor<Double>.getTensor(i: Int): DoubleTensor {
        val dt = asDoubleTensor()
        val lastShape = shape.last(shape.size - 1)
        val newShape: ShapeND = if (lastShape.isNotEmpty()) lastShape else ShapeND(1)
        return DoubleTensor(
            newShape,
            dt.source.view(newShape.linearSize * i, newShape.linearSize)
        )
    }

    /**
     * Creates a tensor of a given shape and fills all elements with a given value.
     *
     * @param value the value to fill the output tensor with.
     * @param shape array of integers defining the shape of the output tensor.
     * @return tensor with the [shape] shape and filled with [value].
     */
    public fun full(value: Double, shape: ShapeND): DoubleTensor {
        checkNotEmptyShape(shape)
        val buffer = Float64Buffer(shape.linearSize) { value }
        return DoubleTensor(shape, buffer)
    }

    /**
     * Returns a tensor with the same shape as `input` filled with [value].
     *
     * @param value the value to fill the output tensor with.
     * @return tensor with the `input` tensor shape and filled with [value].
     */
    public fun fullLike(structureND: StructureND<*>, value: Double): DoubleTensor {
        val shape = structureND.shape
        val buffer = Float64Buffer(structureND.indices.linearSize) { value }
        return DoubleTensor(shape, buffer)
    }

    /**
     * Returns a tensor filled with the scalar value `0.0`, with the shape defined by the variable argument [shape].
     *
     * @param shape array of integers defining the shape of the output tensor.
     * @return tensor filled with the scalar value `0.0`, with the [shape] shape.
     */
    public fun zeros(shape: ShapeND): DoubleTensor = full(0.0, shape)

    /**
     * Returns a tensor filled with the scalar value `0.0`, with the same shape as a given array.
     *
     * @return tensor filled with the scalar value `0.0`, with the same shape as `input` tensor.
     */
    public fun zeroesLike(structureND: StructureND<*>): DoubleTensor = fullLike(structureND, 0.0)

    /**
     * Returns a tensor filled with the scalar value `1.0`, with the shape defined by the variable argument [shape].
     *
     * @param shape array of integers defining the shape of the output tensor.
     * @return tensor filled with the scalar value `1.0`, with the [shape] shape.
     */
    public fun ones(shape: ShapeND): DoubleTensor = full(1.0, shape)

    /**
     * Returns a tensor filled with the scalar value `1.0`, with the same shape as a given array.
     *
     * @return tensor filled with the scalar value `1.0`, with the same shape as `input` tensor.
     */
    public fun onesLike(structureND: StructureND<*>): DoubleTensor = fullLike(structureND, 1.0)

    /**
     * Returns a 2D tensor with shape ([n], [n]), with ones on the diagonal and zeros elsewhere.
     *
     * @param n the number of rows and columns
     * @return a 2-D tensor with ones on the diagonal and zeros elsewhere.
     */
    public fun eye(n: Int): DoubleTensor {
        val shape = ShapeND(n, n)
        val buffer = Float64Buffer(n * n) { 0.0 }
        val res = DoubleTensor(shape, buffer)
        for (i in 0 until n) {
            res[intArrayOf(i, i)] = 1.0
        }
        return res
    }

    override fun Double.plus(arg: StructureND<Double>): DoubleTensor = arg.map { this@plus + it }

    override fun StructureND<Double>.plus(arg: Double): DoubleTensor = map { it + arg }

    override fun StructureND<Double>.plus(arg: StructureND<Double>): DoubleTensor = zip(this, arg) { l, r -> l + r }

    override fun Tensor<Double>.plusAssign(value: Double) {
        mapInPlace { it + value }
    }

    override fun Tensor<Double>.plusAssign(arg: StructureND<Double>) {
        checkShapesCompatible(asDoubleTensor(), arg.asDoubleTensor())
        mapIndexedInPlace { index, value ->
            value + arg[index]
        }
    }

    override fun Double.minus(arg: StructureND<Double>): DoubleTensor = arg.map { this@minus - it }

    override fun StructureND<Double>.minus(arg: Double): DoubleTensor = map { it - arg }

    override fun StructureND<Double>.minus(arg: StructureND<Double>): DoubleTensor = zip(this, arg) { l, r -> l - r }

    override fun Tensor<Double>.minusAssign(value: Double) {
        mapInPlace { it - value }
    }

    override fun Tensor<Double>.minusAssign(arg: StructureND<Double>) {
        checkShapesCompatible(this, arg)
        mapIndexedInPlace { index, value -> value - arg.getDouble(index) }
    }

    override fun Double.times(arg: StructureND<Double>): DoubleTensor = arg.map { this@times * it }

    override fun StructureND<Double>.times(arg: Double): DoubleTensor = arg * asDoubleTensor()

    override fun StructureND<Double>.times(arg: StructureND<Double>): DoubleTensor = zip(this, arg) { l, r -> l * r }

    override fun Tensor<Double>.timesAssign(value: Double) {
        mapInPlace { it * value }
    }

    override fun Tensor<Double>.timesAssign(arg: StructureND<Double>) {
        checkShapesCompatible(this, arg)
        mapIndexedInPlace { index, value -> value * arg[index] }
    }

    override fun Double.div(arg: StructureND<Double>): DoubleTensor = arg.map { this@div / it }

    override fun StructureND<Double>.div(arg: Double): DoubleTensor = map { it / arg }

    override fun StructureND<Double>.div(arg: StructureND<Double>): DoubleTensor = zip(this, arg) { l, r -> l / r }

    override fun Tensor<Double>.divAssign(value: Double) {
        mapInPlace { it / value }
    }

    override fun Tensor<Double>.divAssign(arg: StructureND<Double>) {
        checkShapesCompatible(asDoubleTensor(), arg)
        mapIndexedInPlace { index, value -> value / arg[index] }
    }

    override fun StructureND<Double>.unaryMinus(): DoubleTensor = map { -it }

    override fun StructureND<Double>.transposed(i: Int, j: Int): Tensor<Double> {
        val actualI = if (i >= 0) i else shape.size + i
        val actualJ = if (j >= 0) j else shape.size + j
        return asDoubleTensor().permute(
            shape.transposed(actualI, actualJ)
        ) { originIndex ->
            originIndex.copyOf().apply {
                val ith = get(actualI)
                val jth = get(actualJ)
                set(actualI, jth)
                set(actualJ, ith)
            }
        }
//        // TODO change strides instead of changing content
//        val dt = asDoubleTensor()
//        val ii = dt.minusIndex(i)
//        val jj = dt.minusIndex(j)
//        checkTranspose(dt.dimension, ii, jj)
//        val n = dt.linearSize
//        val resBuffer = DoubleArray(n)
//
//        val resShape = dt.shape.copyOf()
//        resShape[ii] = resShape[jj].also { resShape[jj] = resShape[ii] }
//
//        val resTensor = DoubleTensor(resShape, resBuffer.asBuffer())
//
//        for (offset in 0 until n) {
//            val oldMultiIndex = dt.indices.index(offset)
//            val newMultiIndex = oldMultiIndex.copyOf()
//            newMultiIndex[ii] = newMultiIndex[jj].also { newMultiIndex[jj] = newMultiIndex[ii] }
//
//            val linearIndex = resTensor.indices.offset(newMultiIndex)
//            resTensor.source[linearIndex] = dt.source[offset]
//        }
//        return resTensor
    }

    override fun Tensor<Double>.view(shape: ShapeND): DoubleTensor {
        checkView(asDoubleTensor(), shape)
        return DoubleTensor(shape, asDoubleTensor().source)
    }

    override fun Tensor<Double>.viewAs(other: StructureND<Double>): DoubleTensor =
        view(other.shape)

    /**
     * Broadcasting Matrix product of two tensors.
     *
     * The behavior depends on the dimensionality of the tensors as follows:
     * 1. If both tensors are 1-dimensional, the dot product (scalar) is returned.
     *
     * 2. If both arguments are 2-dimensional, the matrix-matrix product is returned.
     *
     * 3. If the first argument is 1-dimensional and the second argument is 2-dimensional,
     * a 1 is prepended to its dimension for the purpose of the matrix multiply.
     * After the matrix multiply, depending on the implementation the prepended dimension might be removed.
     *
     * 4. If the first argument is 2-dimensional and the second argument is 1-dimensional,
     * the matrix-vector product is returned.
     *
     * 5. If both arguments are at least 1-dimensional and at least one argument is N-dimensional (where N > 2),
     * then a batched matrix multiply is returned. If the first argument is 1-dimensional,
     * a 1 is prepended to its dimension for the purpose of the batched matrix multiply and removed after.
     * If the second argument is 1-dimensional, a 1 is appended to its dimension for the purpose of the batched matrix
     * multiple and removed after.
     * The non-matrix (i.e., batch) dimensions are broadcast (and thus must be broadcastable).
     * For example, if `input` is a (j &times; 1 &times; n &times; n) tensor and `other` is a
     * (k &times; n &times; n) tensor, out will be a (j &times; k &times; n &times; n) tensor.
     *
     * For more information: https://pytorch.org/docs/stable/generated/torch.matmul.html
     *
     * @param other tensor to be multiplied.
     * @return a mathematical product of two tensors.
     */
    public infix fun StructureND<Double>.matmul(other: StructureND<Double>): DoubleTensor {
        if (shape.size == 1 && other.shape.size == 1) {
            return DoubleTensor(ShapeND(1), Float64Buffer(times(other).sum()))
        }

        var penultimateDim = false
        var lastDim = false

        //TODO do we need protective copy here?
        var newThis: DoubleTensor = copyToTensor()
        var newOther: DoubleTensor = other.copyToTensor()

        if (shape.size == 1) {
            penultimateDim = true
            newThis = newThis.view(ShapeND(1) + shape)
        }

        if (other.shape.size == 1) {
            lastDim = true
            newOther = newOther.view(other.shape + intArrayOf(1))
        }

        val broadcastTensors = broadcastOuterTensors(newThis, newOther)
        newThis = broadcastTensors[0]
        newOther = broadcastTensors[1]

        val l = newThis.shape[newThis.shape.size - 2]
        val m1 = newThis.shape[newThis.shape.size - 1]
        val m2 = newOther.shape[newOther.shape.size - 2]
        val n = newOther.shape[newOther.shape.size - 1]
        check(m1 == m2) {
            "Tensors dot operation dimension mismatch: ($l, $m1) x ($m2, $n)"
        }

        val resShape = newThis.shape.slice(0..(newThis.shape.size - 2)) + intArrayOf(newOther.shape.last())
        val resSize = resShape.linearSize
        val resTensor = DoubleTensor(resShape, DoubleArray(resSize).asBuffer())

        val resMatrices = resTensor.matrices
        val newThisMatrices = newThis.matrices
        val newOtherMatrices = newOther.matrices

        for (i in resMatrices.indices) {
            dotTo(newThisMatrices[i], newOtherMatrices[i], resMatrices[i], l, m1, n)
        }
//
//        for ((res, ab) in resTensor.matrixSequence().zip(newThis.matrixSequence().zip(newOther.matrixSequence()))) {
//            val (a, b) = ab
//            dotTo(a, b, res, l, m1, n)
//        }

        return if (penultimateDim) {
            resTensor.view(resTensor.shape.first(resTensor.shape.size - 2) + ShapeND(resTensor.shape.last()))
        } else if (lastDim) {
            resTensor.view(resTensor.shape.first(resTensor.shape.size - 1))
        } else {
            resTensor
        }
    }

    override fun StructureND<Double>.dot(other: StructureND<Double>): DoubleTensor {
        return if (dimension in 0..2 && other.dimension in 0..2) this.matmul(other)
        else error("Only vectors and matrices are allowed in non-broadcasting dot operation")
    }

    override fun diagonalEmbedding(
        diagonalEntries: StructureND<Double>,
        offset: Int,
        dim1: Int,
        dim2: Int,
    ): DoubleTensor {
        val n = diagonalEntries.shape.size
        val d1 = minusIndexFrom(n + 1, dim1)
        val d2 = minusIndexFrom(n + 1, dim2)

        check(d1 != d2) {
            "Diagonal dimensions cannot be identical $d1, $d2"
        }
        check(d1 <= n && d2 <= n) {
            "Dimension out of range"
        }

        var lessDim = d1
        var greaterDim = d2
        var realOffset = offset
        if (lessDim > greaterDim) {
            realOffset *= -1
            lessDim = greaterDim.also { greaterDim = lessDim }
        }

        val resShape = diagonalEntries.shape.slice(0 until lessDim) +
                intArrayOf(diagonalEntries.shape[n - 1] + abs(realOffset)) +
                diagonalEntries.shape.slice(lessDim until greaterDim - 1) +
                intArrayOf(diagonalEntries.shape[n - 1] + abs(realOffset)) +
                diagonalEntries.shape.slice(greaterDim - 1 until n - 1)
        val resTensor: DoubleTensor = zeros(resShape)

        for (i in 0 until diagonalEntries.indices.linearSize) {
            val multiIndex = diagonalEntries.indices.index(i)

            var offset1 = 0
            var offset2 = abs(realOffset)
            if (realOffset < 0) {
                offset1 = offset2.also { offset2 = offset1 }
            }
            val diagonalMultiIndex = multiIndex.slice(0 until lessDim).toIntArray() +
                    intArrayOf(multiIndex[n - 1] + offset1) +
                    multiIndex.slice(lessDim until greaterDim - 1).toIntArray() +
                    intArrayOf(multiIndex[n - 1] + offset2) +
                    multiIndex.slice(greaterDim - 1 until n - 1).toIntArray()

            resTensor[diagonalMultiIndex] = diagonalEntries[multiIndex]
        }

        return resTensor
    }

    /**
     * Compares element-wise two tensors with a specified precision.
     *
     * @param other the tensor to compare with `input` tensor.
     * @param epsilon permissible error when comparing two Double values.
     * @return true if two tensors have the same shape and elements, false otherwise.
     */
    public fun Tensor<Double>.eq(other: Tensor<Double>, epsilon: Double): Boolean =
        asDoubleTensor().eq(other) { x, y -> abs(x - y) < epsilon }

    /**
     * Compares element-wise two tensors.
     * Comparison of two Double values occurs with `1e-5` precision.
     *
     * @param other the tensor to compare with `input` tensor.
     * @return true if two tensors have the same shape and elements, false otherwise.
     */
    public infix fun Tensor<Double>.eq(other: Tensor<Double>): Boolean = eq(other, 1e-5)

    private fun Tensor<Double>.eq(
        other: Tensor<Double>,
        eqFunction: (Double, Double) -> Boolean,
    ): Boolean {
        //TODO optimize tensor conversion
        checkShapesCompatible(asDoubleTensor(), other)
        val n = asDoubleTensor().linearSize
        if (n != other.asDoubleTensor().linearSize) {
            return false
        }
        for (i in 0 until n) {
            if (!eqFunction(asDoubleTensor().source[i], other.asDoubleTensor().source[i])) {
                return false
            }
        }
        return true
    }

    /**
     * Builds tensor from rows of the input tensor.
     *
     * @param indices the [IntArray] of 1-dimensional indices
     * @return tensor with rows corresponding to row by [indices]
     */
    public fun Tensor<Double>.rowsByIndices(indices: IntArray): DoubleTensor = stack(indices.map { getTensor(it) })


    private inline fun StructureND<Double>.foldDimToDouble(
        dim: Int,
        keepDim: Boolean,
        foldFunction: (DoubleArray) -> Double,
    ): DoubleTensor {
        check(dim < dimension) { "Dimension $dim out of range $dimension" }
        val resShape = if (keepDim) {
            shape.first(dim) + intArrayOf(1) + shape.last(dimension - dim - 1)
        } else {
            shape.first(dim) + shape.last(dimension - dim - 1)
        }
        val resNumElements = resShape.linearSize
        val init = foldFunction(DoubleArray(1) { 0.0 })
        val resTensor = DoubleTensor(
            resShape,
            Float64Buffer(resNumElements) { init }
        )
        val dt = asDoubleTensor()
        for (index in resTensor.indices) {
            val prefix = index.take(dim).toIntArray()
            val suffix = index.takeLast(dimension - dim - 1).toIntArray()
            resTensor[index] = foldFunction(DoubleArray(shape[dim]) { i ->
                dt[prefix + intArrayOf(i) + suffix]
            })
        }
        return resTensor
    }

    private inline fun StructureND<Double>.foldDimToInt(
        dim: Int,
        keepDim: Boolean,
        foldFunction: (DoubleArray) -> Int,
    ): IntTensor {
        check(dim < dimension) { "Dimension $dim out of range $dimension" }
        val resShape = if (keepDim) {
            shape.first(dim) + intArrayOf(1) + shape.last(dimension - dim - 1)
        } else {
            shape.first(dim) + shape.last(dimension - dim - 1)
        }
        val resNumElements = resShape.linearSize
        val init = foldFunction(DoubleArray(1) { 0.0 })
        val resTensor = IntTensor(
            resShape,
            Int32Buffer(resNumElements) { init }
        )
        for (index in resTensor.indices) {
            val prefix = index.take(dim).toIntArray()
            val suffix = index.takeLast(dimension - dim - 1).toIntArray()
            resTensor[index] = foldFunction(DoubleArray(shape[dim]) { i ->
                asDoubleTensor()[prefix + intArrayOf(i) + suffix]
            })
        }
        return resTensor
    }


    override fun StructureND<Double>.sum(): Double = reduceElements { it.array.sum() }

    override fun StructureND<Double>.sum(dim: Int, keepDim: Boolean): DoubleTensor =
        foldDimToDouble(dim, keepDim) { x -> x.sum() }

    override fun StructureND<Double>.min(): Double = reduceElements { it.array.min() }

    override fun StructureND<Double>.min(dim: Int, keepDim: Boolean): DoubleTensor =
        foldDimToDouble(dim, keepDim) { x -> x.minOrNull()!! }

    override fun StructureND<Double>.argMin(dim: Int, keepDim: Boolean): Tensor<Int> = foldDimToInt(dim, keepDim) { x ->
        x.withIndex().minBy { it.value }.index
    }

    override fun StructureND<Double>.max(): Double = reduceElements { it.array.max() }

    override fun StructureND<Double>.max(dim: Int, keepDim: Boolean): DoubleTensor =
        foldDimToDouble(dim, keepDim) { x -> x.maxOrNull()!! }


    override fun StructureND<Double>.argMax(dim: Int, keepDim: Boolean): IntTensor =
        foldDimToInt(dim, keepDim) { x ->
            x.withIndex().maxBy { it.value }.index
        }


    override fun mean(structureND: StructureND<Double>): Double = structureND.sum() / structureND.indices.linearSize

    override fun mean(structureND: StructureND<Double>, dim: Int, keepDim: Boolean): Tensor<Double> =
        structureND.foldDimToDouble(dim, keepDim) { arr ->
            check(dim < structureND.dimension) { "Dimension $dim out of range ${structureND.dimension}" }
            arr.sum() / structureND.shape[dim]
        }

    override fun std(structureND: StructureND<Double>): Double = structureND.reduceElements { arr ->
        val mean = arr.array.sum() / structureND.indices.linearSize
        sqrt(arr.array.sumOf { (it - mean) * (it - mean) } / (structureND.indices.linearSize - 1))
    }

    override fun std(structureND: StructureND<Double>, dim: Int, keepDim: Boolean): Tensor<Double> =
        structureND.foldDimToDouble(
            dim,
            keepDim
        ) { arr ->
            check(dim < structureND.dimension) { "Dimension $dim out of range ${structureND.dimension}" }
            val mean = arr.sum() / structureND.shape[dim]
            sqrt(arr.sumOf { (it - mean) * (it - mean) } / (structureND.shape[dim] - 1))
        }

    override fun variance(structureND: StructureND<Double>): Double = structureND.reduceElements { arr ->
        val linearSize = structureND.indices.linearSize
        val mean = arr.array.sum() / linearSize
        arr.array.sumOf { (it - mean) * (it - mean) } / (linearSize - 1)
    }

    override fun variance(structureND: StructureND<Double>, dim: Int, keepDim: Boolean): Tensor<Double> =
        structureND.foldDimToDouble(
            dim,
            keepDim
        ) { arr ->
            check(dim < structureND.dimension) { "Dimension $dim out of range ${structureND.dimension}" }
            val mean = arr.sum() / structureND.shape[dim]
            arr.sumOf { (it - mean) * (it - mean) } / (structureND.shape[dim] - 1)
        }


    override fun exp(arg: StructureND<Double>): DoubleTensor = arg.map { this.exp(it) }

    override fun ln(arg: StructureND<Double>): DoubleTensor = arg.map { this.ln(it) }

    override fun sqrt(arg: StructureND<Double>): DoubleTensor = arg.map { this.sqrt(it) }

    override fun cos(arg: StructureND<Double>): DoubleTensor = arg.map { this.cos(it) }

    override fun acos(arg: StructureND<Double>): DoubleTensor = arg.map { this.acos(it) }

    override fun cosh(arg: StructureND<Double>): DoubleTensor = arg.map { this.cosh(it) }

    override fun acosh(arg: StructureND<Double>): DoubleTensor = arg.map { this.acosh(it) }

    override fun sin(arg: StructureND<Double>): DoubleTensor = arg.map { this.sin(it) }

    override fun asin(arg: StructureND<Double>): DoubleTensor = arg.map { this.asin(it) }

    override fun sinh(arg: StructureND<Double>): DoubleTensor = arg.map { this.sinh(it) }

    override fun asinh(arg: StructureND<Double>): DoubleTensor = arg.map { this.asinh(it) }

    override fun tan(arg: StructureND<Double>): DoubleTensor = arg.map { this.tan(it) }

    override fun atan(arg: StructureND<Double>): DoubleTensor = arg.map { this.atan(it) }

    override fun tanh(arg: StructureND<Double>): DoubleTensor = arg.map { this.tanh(it) }

    override fun atanh(arg: StructureND<Double>): DoubleTensor = arg.map { this.atanh(it) }

    override fun power(arg: StructureND<Double>, pow: Number): StructureND<Double> = if (pow is Int) {
        arg.map { it.pow(pow) }
    } else {
        arg.map { it.pow(pow.toDouble()) }
    }

    override fun ceil(arg: StructureND<Double>): DoubleTensor = arg.map { ceil(it) }

    override fun floor(structureND: StructureND<Double>): DoubleTensor = structureND.map { floor(it) }

    override fun StructureND<Double>.inv(): DoubleTensor = invLU(this, 1e-9)

    override fun StructureND<Double>.det(): DoubleTensor = detLU(this, 1e-9)

    override fun lu(structureND: StructureND<Double>): Triple<DoubleTensor, DoubleTensor, DoubleTensor> =
        lu(structureND, 1e-9)

    override fun cholesky(structureND: StructureND<Double>): DoubleTensor = cholesky(structureND, 1e-6)

    override fun qr(structureND: StructureND<Double>): Pair<DoubleTensor, DoubleTensor> {
        checkSquareMatrix(structureND.shape)
        val qTensor = zeroesLike(structureND)
        val rTensor = zeroesLike(structureND)

        //TODO replace with cycle
        structureND.asDoubleTensor().matrixSequence()
            .zip(
                (qTensor.matrixSequence()
                    .zip(rTensor.matrixSequence()))
            ).forEach { (matrix, qr) ->
                val (q, r) = qr
                qrHelper(matrix, q, r.asDoubleTensor2D())
            }

        return qTensor to rTensor
    }

    override fun svd(
        structureND: StructureND<Double>,
    ): Triple<StructureND<Double>, StructureND<Double>, StructureND<Double>> =
        svdGolubKahan(structureND = structureND, epsilon = 1e-10)

    override fun symEig(structureND: StructureND<Double>): Pair<DoubleTensor, DoubleTensor> =
        symEigJacobi(structureND = structureND, maxIteration = 50, epsilon = 1e-15)

    override fun solve(a: MutableStructure2D<Double>, b: MutableStructure2D<Double>): MutableStructure2D<Double> {
        val aSvd = DoubleTensorAlgebra.svd(a)
        val s = BroadcastDoubleTensorAlgebra.diagonalEmbedding(aSvd.second.map { 1.0 / it })
        val aInverse = aSvd.third.dot(s).dot(aSvd.first.transposed())
        return aInverse.dot(b).as2D()
    }
}

public val Double.Companion.tensorAlgebra: DoubleTensorAlgebra get() = DoubleTensorAlgebra
public val Float64Field.tensorAlgebra: DoubleTensorAlgebra get() = DoubleTensorAlgebra